/*-------------------------------------------------------------------*/
/* PNNFAST2.C          Pasi Frnti & Olli Virmajoki                   */
/*                                                                   */
/* Fast PNN method with PDS & MPS & Lazy                             */
/*   Starts from initial solution                                    */
/*	 Reduces cluster size to FinalSize								 */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName	   "PNN-FAST"
#define VersionNumber  "Version 0.21"
#define LastUpdated    "16.9.2000"

/* ----------------------------------------------------------------- */

#include <assert.h>
#include <float.h>
#include <stdio.h>
#include <time.h>
#include <values.h>

#include "cb.h"
#include "interfc.h"
#include "memctrl.h"
#include "random.h"
#include "sort.h"
#include "solution.h"
#include "pnnfast.h"


/*---------------------------  M i s c	------------------------------*/


/*#define  STATISTICS 1*/
#ifdef STATISTICS
#define STATCODE(a)a
#else 
#define STATCODE(a)
#endif

#define  round(a)		((a)+0.5)
#define  roundpds(a)    ((a)+0.9999999)
#define  swap(a,b)      { int t = a; a = b; b = t; }


llong ksum=0LL;
llong ksumtotal=0LL;
llong ndistcalcs=0LL;
llong ndistcalcstotal=0LL;
llong nnccalcs=0LL;
llong nnccalcstotal=0LL;
llong ntau=0LL;
llong ntautotal=0LL;
llong nterm=0LL;
llong npre=0LL;
llong npretotal=0LL;
llong niteration=0LL;


static void FindNearestClusterMPS(CODEBOOK* CB, DATA* D, int a,
                                  int* b, llong* dist);


/*============================  P R I N T  ===============================*/


static void PrintTime(long start)
{
  time_t tmp;
  long t = (long)time(&tmp) - start;
  long h,m,s;

  s=t%60; t=t/60; m=t%60; h=t/60;
  printf("  %li:%02li:%02li\n", h, m, s);
  fflush(stdout);
}


/*-------------------------------------------------------------------*/


static void PrintProgress(TRAININGSET*   TS,
                          CODEBOOK*      CB,
                          PARTITIONING*  P,
                          int            Quiet,
                          long           watch)
{
  double Error;

  if(Quiet<2) return;
  ksumtotal += ksum;
  ndistcalcstotal += ndistcalcs;
  nnccalcstotal += nnccalcs;
  ntautotal += ntau;
  npretotal += npre;
  niteration += 1;
  Error = AverageErrorForSolution(TS, CB, P, MSE);
  Error = PrintableError(Error, CB);

  switch(Quiet)
  {

/*    case 2:  
        printf("\n%3i: %9.4f\ntau: %7.2f term: %7.2f\ndist calcs/search: %7.2f dim/dist calcs: %7.2f dim/seaerch: %7.2f ksum: %7.0f ndistcalcs: %7.0f searches: %7.0f",
        BookSize(CB), Error,
        (double)ntau, (double)nterm,
        nnccalcs != 0LL ? (double)ndistcalcs/nnccalcs : 0.0,
        ndistcalcs != 0LL ? (double)ksum/ndistcalcs : 0.0,
        nnccalcs != 0LL ? (double)ksum/nnccalcs : 0.0,
        (double)ksum,(double)ndistcalcs,(double)nnccalcs);
        break;
    case 3:  
        printf("%3i: %9.4f ",  BookSize(CB), Error); PrintTime(watch); break;
    default: 
        printf("Book size = %3i   Error=%9.4f  Time=", BookSize(CB), Error);
             PrintTime(watch); break;
*/  }
  fflush(stdout);
  ksum=0LL;
  ndistcalcs=0LL;
  nnccalcs=0LL;
  ntau=0LL;
  npre=0LL;
  nterm=0LL;
}


/*====================================================================*/


static llong MergeDistortionORI(CODEBOOK* CB, int p1, int p2)
{
  /* Original */
  llong   dist;
  double  factor;

  dist = VectorDist(Vector(CB,p1), Vector(CB,p2), VectorSize(CB));
  STATCODE
      (
      ksum+=VectorSize(CB); /* Statistical information */
      )
  factor = ((double)VectorFreq(CB,p1) * (double)VectorFreq(CB,p2)) /
           ((double)VectorFreq(CB,p1) + (double)VectorFreq(CB,p2));
  dist = round( factor * (double)dist);
  STATCODE
      (
      ndistcalcs++; /* Statistical information */
      )
  return(dist);
}


static llong MergeDistortion(CODEBOOK* CB, int p1, int p2)
{
  /* Original "inline" */
  int i=0, Vsize;
  llong   dist=0LL, diff, dist_limit=MAXLLONG;
  double  factor;

  Vsize = VectorSize(CB);
  factor = ((double)VectorFreq(CB,p1) * (double)VectorFreq(CB,p2)) /
           ((double)VectorFreq(CB,p1) + (double)VectorFreq(CB,p2));
	 /* Precondition: Vsize >= 1 */
     /* if dist >= dist_limit, returns at least factor * dist_limit */
  do
	{
	   diff = (llong)VectorScalar(CB,p1,i)-(llong)VectorScalar(CB,p2,i);
	   dist += diff * diff;
       STATCODE
           (
           ksum++; /* Statistical information */
           )
	} while(dist < dist_limit && ++i < Vsize);
  dist=round( factor * (double)dist);
  STATCODE
      (
      ndistcalcs++; /* Statistical information */
      )

  return( dist);
}


/*-------------------------------------------------------------------*/


static llong MergeDistortionPdsMPS(CODEBOOK* CB, int p1, int p2,
								llong dist_limit)
{
  /* Variant Olli b */
  int i=0, Vsize;
  llong   dist=0LL, diff, distortion=0LL;
  double  factor;

  Vsize = VectorSize(CB);
  factor = ((double)VectorFreq(CB,p1) * (double)VectorFreq(CB,p2)) /
           ((double)VectorFreq(CB,p1) + (double)VectorFreq(CB,p2));
  /* Precondition: Vsize >= 1 */
  /* if distortion >= dist_limit, returns at least dist_limit */
  for(i=0;i<Vsize;i++)	
	{
	   diff = (llong)VectorScalar(CB,p1,i)-(llong)VectorScalar(CB,p2,i);
	   dist += diff * diff;
       STATCODE
           (
    	   ksum++; /* Statistical information */
           )
	   distortion=round( factor * (double)dist);
	   if(distortion > dist_limit) return(dist_limit);
	}
    STATCODE
        (
        ndistcalcs++; /* Statistical information */
        )
  return( distortion);
}


static llong MergeDistortionPdsSIM(CODEBOOK* CB, int p1, int p2,
								llong dist_limit)
{
  /* Variant Simple */
  int i=0, Vsize;
  llong   dist=0LL, diff, distortion=0LL;
  double  factor;

  Vsize = VectorSize(CB);
  factor = ((double)VectorFreq(CB,p1) * (double)VectorFreq(CB,p2)) /
           ((double)VectorFreq(CB,p1) + (double)VectorFreq(CB,p2));
  /* Precondition: Vsize >= 1 */
  /* if distortion >= dist_limit, returns at least dist_limit */
  do	
	{
	   diff = (llong)VectorScalar(CB,p1,i)-(llong)VectorScalar(CB,p2,i);
	   dist += diff * diff;
       STATCODE
           (
           ksum++; /* Statistical information */
           )
	   distortion=round( factor * (double)dist);
	} while( distortion < dist_limit && ++i < Vsize);
    STATCODE
        (
        ndistcalcs++; /* Statistical information */
        )
  return( distortion);
}


static llong MergeDistortionPds(CODEBOOK* CB, int p1, int p2,
								  llong maxdist)
{
  /* Variant Optimized */
  int i=0, Vsize;
  llong   dist=0LL, diff, distortion=0LL, dist_limit;
  double  factor;

  Vsize = VectorSize(CB);
  factor = ((double)VectorFreq(CB,p1) * (double)VectorFreq(CB,p2)) /
           ((double)VectorFreq(CB,p1) + (double)VectorFreq(CB,p2));
  dist_limit = roundpds((double)maxdist/factor);
  /* Precondition: Vsize >= 1 */
  /* if dist >= dist_limit, returns at least maxdist */
  do
	 {
		diff = (llong)VectorScalar(CB,p1,i)-(llong)VectorScalar(CB,p2,i);
		dist += diff * diff;
        STATCODE
            (
    		ksum++; /* Statistical information */
            )
	 } while(dist < dist_limit && ++i < Vsize);
  distortion=round( factor * (double)dist);
  STATCODE
      (
      ndistcalcs++; /* Statistical information */
      )
  return( distortion);
}


/*-------------------------------------------------------------------*/


static void FindNearestCluster(CODEBOOK* CB, int a, int* b,
								  llong* dist, int Pds)
{
  llong Dist;
  int i, aex;

  if(Pds)
  {
    aex =(*b);
    if( a!=(*b))
    {
        (*dist)=MAXLLONG/10;
	    (*dist)=MergeDistortionPds(CB, a, *b, *dist);
    }
    else
    {
	    (*dist)=MAXLLONG/10; 
	    (*b)=0;
    }
    for( i=0; i<BookSize(CB); i++ )
    {
    if( a!=i && aex!=i)
	    {
        Dist = MergeDistortionPds(CB, a, i, *dist);
        if( Dist < (*dist) )
		   {
              (*dist) = Dist;
              (*b)    = i;
           }
	    }
    }
  }
  else
  {
    (*dist)=MAXLLONG/10; (*b)=0;
    for( i=0; i<BookSize(CB); i++ )
    {
        if( a!=i)
	    {
		    Dist = MergeDistortion(CB, a, i);
            if( Dist < (*dist) )
            {
                (*dist) = Dist;
                (*b)    = i;
            }
        }
    }
  }
}


/*-------------------------------------------------------------------*/


void CalculateAllDistances(CODEBOOK* CB, DATA* D, NEARESTTYPE NN[], int Quiet, 
                                  long watch, int Pds, int Mps)
/* vikkel / Timos variant */
{
  int  i,j;
  llong Dist;

  if(Quiet>=3) printf("Calculating distance matrix:\n"); fflush(stdout);

  if(Mps)
  {
  for( i=0; i<BookSize(CB); i++)
    {
        NN[i].Vector = i;
	    FindNearestClusterMPS(CB, D, i,
            &(NN[i].Vector), &(NN[i].VectorDist));
        NN[i].Recalculate = NO;
        if(ExitRequested()) ExitProcessing(-2);
    }
  }
  else
  {
  for( i=0; i<BookSize(CB)-1; i++)
	 for(j=i+1; j<BookSize(CB);j++)
     {
		  if(Pds)
		  {
             Dist=MergeDistortionPds(CB,i,j,NN[i].VectorDist);
		  }
		  else
		  {
			 Dist=MergeDistortion(CB,i,j);
		  }
		  if(Dist<NN[i].VectorDist)
		  {
			 NN[i].VectorDist=Dist;
			 NN[i].Vector=j;
		  }
     if(ExitRequested()) ExitProcessing(-2);
     }
  }
  STATCODE
      (
      nnccalcs=BookSize(CB);
      )
    if(Quiet>=3) { printf("\r                    \r"); PrintTime(watch); }
}


/*-------------------------------------------------------------------*/


static void FindMergedPair(CODEBOOK* CB, DATA* D, NEARESTTYPE NN[], int* a, int* b,
                           int Pds, int Mps)
{
  llong MinDist=MAXLLONG;
  int i;
  int nrec=0;

  for( i=0; i<BookSize(CB); i++ )
    {
       if( NN[i].VectorDist < MinDist )
	   {
          if(NN[i].Recalculate)
          {
              if(Mps)
              {
                FindNearestClusterMPS(CB, D, i,
			        &(NN[i].Vector), &(NN[i].VectorDist));
              }
              else
              FindNearestCluster(CB, i, &(NN[i].Vector), 
                &(NN[i].VectorDist), Pds);
              NN[i].Recalculate = NO;
              nrec++;
              if(NN[i].VectorDist < MinDist)
              {
                (*a) = i;
                (*b) = NN[i].Vector;
                MinDist = NN[i].VectorDist;
              }
          }
          else
          {
            (*a) = i;
            (*b) = NN[i].Vector;
            MinDist = NN[i].VectorDist;
          }
       }
    }
  if( (*a)>(*b) ) swap((*a),(*b));
  STATCODE
      (
      ntau=nrec;
      nnccalcs=nrec;
      )
/*  if(NN[*a].VectorDist != NN[*b].VectorDist)
	{
		PrintMessage("Warning: Distance of minimum pair is not identical!\n");
	}*/
}


/*-------------------------------------------------------------------*/


static void CentroidOfTwoVectors(CODEBOOK* CB, BOOKNODE* cent, int p1, int p2)
{
  int k;
  int freqsum;

  freqsum = VectorFreq(CB,p1) + VectorFreq(CB,p2);

  /* Calculate mean values for centroid */
  for( k=0; k<VectorSize(CB); k++)
    {
	if( freqsum > 0 )
	cent->vector[k] = round( (double)(VectorFreq(CB, p1)*VectorScalar(CB, p1, k) +
                                      VectorFreq(CB, p2)*VectorScalar(CB, p2, k))
                             / (double)freqsum );
	else cent->vector[k] = 0;
    }
  cent->freq = freqsum;
}


/* ----------------------------------------------------------------- */


static void RemoveLastVector(CODEBOOK* CB, int a, int b, NEARESTTYPE NN[])
{
  int last = BookSize(CB)-1;
  int i;

  if( b != last )
    {
    CopyNode(&Node(CB,last), &Node(CB,b), VectorSize(CB));
    NN[b] = NN[last];
    for( i=0; i<BookSize(CB); i++ )
       {
       if( NN[i].Vector==last )  NN[i].Vector = b;
       }
    }
}


/*--------------------------------------------------------------------*/


void RemoveEmptyPartitions(TRAININGSET*  TS,
                           CODEBOOK*     CB,
                           PARTITIONING* P,
                           int           FinalSize)
{
  int i,last;

  for( i=0; i<BookSize(CB); i++ )
     {
     if( VectorFreq(CB,i)==0  &&  BookSize(CB)>FinalSize )
        {
        last = BookSize(CB)-1;
        CopyNode(&Node(CB,last), &Node(CB,i), VectorSize(CB));
        JoinPartitions(TS, P, i, last);
        BookSize(CB)--;
        i--;
        }
     }
}


/*--------------------------------------------------------------------*/


static void MarkClustersForRecalculation(CODEBOOK* CB,
                                         NEARESTTYPE NN[],
                                         int a, int b)
{
  int i;
  
  for( i=0; i<BookSize(CB); i++ )
    {
    if(NN[i].Vector==a || NN[i].Vector==b)
        NN[i].Recalculate = YES;
	}
  NN[a].Recalculate = YES;
}


/*--------------------------------------------------------------------*/


void RecalculateDistances(CODEBOOK* CB, NEARESTTYPE NN[], int aex,
								 int Quiet, int Pds, int Lazy)
{
  int  i;
  int  nrec=0;

  if(Lazy)
  {
     FindNearestCluster(CB, aex, &(NN[aex].Vector), &(NN[aex].VectorDist), Pds);
     NN[aex].Recalculate = NO;
     nrec++;
  }
  else
  {
  for( i=0; i<BookSize(CB); i++ )
    {
    if( NN[i].Recalculate )
       {
	   FindNearestCluster(CB, i, &(NN[i].Vector), &(NN[i].VectorDist), Pds);
       NN[i].Recalculate = NO;
       nrec++;
       }
    }
  }
  STATCODE
      (
      nnccalcs+=nrec;
      ntau+=nrec;
      )
  if(Quiet>=4) printf("Recalculated: %i / %i.\n", nrec, BookSize(CB));
}


/*-------------------------------------------------------------------*/


void MergeVectors(TRAININGSET*  TS,
                  CODEBOOK*     CB,
                  PARTITIONING* P,
                  NEARESTTYPE   NN[],
                  int           a,
                  int           b)
{
  int i;
  MarkClustersForRecalculation(CB, NN, a, b);
/*  CentroidOfTwoVectors(CB, &Node(CB, a), a, b);*/   
  JoinPartitions(TS, P, a, b);
  PartitionCentroid(P, a, &Node(CB, a));
  for( i=0; i<BookSize(CB); i++)
  {
      if(NN[i].Vector==b) NN[i].Vector=a;
  }
  RemoveLastVector(CB, a, b, NN);
  BookSize(CB)--;
}


/* ========================  MPS  ================================== */


/*-------------------------------------------------------------------*/


static int VectorSum(int* vector, int dim)
{
  int sum = 0;
  int k;

  for( k = 0; k < dim; k++ )
    {
    sum += vector[k];
    }
  return( sum );
}


/*-------------------------------------------------------------------*/


static void CalculateCodevectorSums(CODEBOOK* CB, DATA* D)
{
  int i;

  for( i = 0; i < BookSize(CB); i++ )
    {
    D->codesum[i] = VectorSum(Vector(CB, i), VectorSize(CB));
    }
}


/*-------------------------------------------------------------------*/


static void RecalculateCodevectorSum(CODEBOOK* CB, DATA* D, int a, int b)
{
  int last = BookSize(CB);

  D->codesum[a] = VectorSum(Vector(CB, a), VectorSize(CB));
  D->codesum[b] = D->codesum[last];
}


/*-------------------------------------------------------------------*/


static llong SMD(int sum1, int sum2)
{
  return( (llong)(sum1 - sum2) * (llong)(sum1 - sum2) );
}


/*-------------------------------------------------------------------*/


static void FindNearestClusterMPS(CODEBOOK* CB, DATA* D, int a,
                                  int* b, llong* dist)
{
  int i,aex;
  llong Dist;
  
  aex = *b;
  if(a !=(*b))
  {
     (*dist) = MAXLLONG/VectorSize(CB);
     (*dist) = MergeDistortionPds(CB, a, *b, *dist);
  }
  else
  {
     (*dist) = MAXLLONG/VectorSize(CB);
     (*b)    = 0;
  }
  for( i=0; i < BookSize(CB); i++)
  {
     if( a!=i && aex!=i)
     {
         STATCODE
                 (
                 npre += 1;
                 )
         if(
		        (((double)VectorFreq(CB,a) * (double)VectorFreq(CB,i) /
                ((double)VectorFreq(CB,a) + (double)VectorFreq(CB,i))) * 
		        SMD(D->codesum[a], D->codesum[i]))
		         > VectorSize(CB) * (*dist)
	       )
         {
             
             continue;
         }
         else
         {
            Dist=MergeDistortionPds(CB, a, i, *dist);
            if(Dist < (*dist) )
            {
               (*dist) = Dist;
               (*b)    = i;
            }
         }
     }
  }       
}


/*-------------------------------------------------------------------*/


static void RecalculateDistancesMPS(CODEBOOK* CB, DATA* D, 
                                    NEARESTTYPE NN[], int aex,
                                    int Quiet, int Pds, int Lazy)
{
  int  i;
  int  nrec=0;

  if(Lazy)
  {
    FindNearestClusterMPS(CB, D, aex, 
        &(NN[aex].Vector), &(NN[aex].VectorDist));
    NN[aex].Recalculate = NO;
    nrec++;
  }
  else
  {
    for( i=0; i<BookSize(CB); i++)
    {
      if( NN[i].Recalculate )
      {
	     FindNearestClusterMPS(CB, D, i,
			   &(NN[i].Vector), &(NN[i].VectorDist));
         NN[i].Recalculate = NO;
         nrec++;
      }
    }
  }
  STATCODE
      (
      nnccalcs+=nrec;
      ntau+=nrec;
      )
  if(Quiet>=4) printf("Recalculated: %i / %i.\n", nrec, BookSize(CB));
}


/* ================================================================= */


void CreateDataStructures(CODEBOOK* CB, DATA* D, NEARESTTYPE** NN, int Mps)
{
    if(Mps)
    {
    D->codesum = allocate(BookSize(CB) * sizeof(int));
	D->codeorder = allocate(BookSize(CB) * sizeof(int));
	D->mynumber = allocate(BookSize(CB) * sizeof(int));
    }
    *NN = (NEARESTTYPE*) allocate(BookSize(CB) * sizeof(NEARESTTYPE));
}


void InitializeData(CODEBOOK* CB, DATA* D, NEARESTTYPE NN[], int Mps)
{
    int i;
    if(Mps)
    {
	for(i=0; i<BookSize(CB); i++)
    {
		D->codesum[i] = 0;
		D->codeorder[i] = i;
    }
    CalculateCodevectorSums(CB, D);
/*    QuickSort(D->codeorder, BookSize(CB), sizeof(int), D->codesum, cmpCodeSum);
    for(i=0; i<BookSize(CB); i++)
    {
       j=D->codeorder[i];
	   D->mynumber[j]=i;
    }  */
    }
    else
    {
       for(i=0; i<BookSize(CB);i++)
       {
	      NN[i].VectorDist=MAXLLONG/10;
	      NN[i].Vector=i;
          NN[i].Recalculate=NO;
       }
    }
}


static void PrintResult(int Quiet)
{
  if(Quiet>=2)
  {
        printf("\nDone preconditions/search: %7.2f",nnccalcstotal != 0LL ? (double)npretotal/nnccalcstotal : 0.0);
        printf("\ntauAverage: %7.2f \ndist calcs/search: %7.2f dim/dist calcs: %7.2f dim/seaerch: %7.2f ksum: %7.0f ndistcalcs: %7.0f searches: %7.0f",
        (double)ntautotal/niteration,
        nnccalcstotal != 0LL ? (double)ndistcalcstotal/nnccalcstotal : 0.0,
        ndistcalcstotal != 0LL ? (double)ksumtotal/ndistcalcstotal : 0.0,
        nnccalcstotal != 0LL ? (double)ksumtotal/nnccalcstotal : 0.0,
        (double)ksumtotal,(double)ndistcalcstotal,(double)nnccalcstotal);
  }
}


void FreeDataStructures(DATA* D, NEARESTTYPE NN[], int Mps)
{
    if(Mps)
    {
    deallocate(D->codesum);
	deallocate(D->codeorder);
	deallocate(D->mynumber);
    }
    deallocate(NN);
}



/*-------------------------------------------------------------------*/


void PNNFast(TRAININGSET*	TS,
			 CODEBOOK*		CB,
			 PARTITIONING*	P,
             int            FinalSize,
             int            Quiet,
             long           watch,
			 int			Pds,
			 int			Mps,
             int            Lazy)
{
  NEARESTTYPE*  NN;
  DATA D;
  int  a,b;

  RemoveEmptyPartitions(TS, CB, P, FinalSize);
  CreateDataStructures(CB, &D, &NN, FinalSize);
  InitializeData(CB, &D, NN, Mps);
  CalculateAllDistances(CB, &D, NN, Quiet, watch, Pds, Mps);
  /*PrintTime(watch);*/
  PrintProgress(TS, CB, P, Quiet, watch);
  niteration=0LL;

  while( BookSize(CB) > FinalSize)
	{
	  FindMergedPair(CB, &D, NN, &a, &b, Pds, Mps);
	  MergeVectors(TS, CB, P, NN, a, b);
      if(Mps)
        {
        RecalculateCodevectorSum(CB, &D, a, b);
	    RecalculateDistancesMPS(CB, &D, NN, a, Quiet, Pds, Lazy);
        }
      else
        {
        RecalculateDistances(CB, NN, a, Quiet, Pds, Lazy);
        }
      PrintProgress(TS, CB, P, Quiet, watch);
	  if(ExitRequested()) ExitProcessing(-2);
    }

  PrintResult(Quiet);
  FreeDataStructures(&D, NN, Mps);
}

